#include <stdio.h>
#include <p4.h>

/* #define VERBOSE */
/* #define COUNT */
/* #define REPORT_COUNT */

#define BUFFER 2

#define ANY -1
#define SETUP 1
#define WORK 2
#define RESULT 3
#define DATA 4

#define maxdepth 64
#define ratio 1.462

void area();
void parcalc();
void header();
void footer();
void test(); 
void switches();

FILE* out;
char* outfile = "m.xbm";
char str[maxdepth], prev[maxdepth];
int iter, user_iter = -1, autoiter[maxdepth];
double a,b,as,bs;
int Xmax=-1,Ymax=-1,nswitches=0;
int verbose = 0;
int timing = 0;
int stage = 15;

void setupvars() {
  int i;
  char *s;
  for (i=0 ; i<=maxdepth; i++) 
    autoiter[i]=40+10*i;
  if (Xmax== -1 && Ymax== -1) Xmax=64;
  if (Xmax != -1) Xmax = (Xmax + stage) & ~stage; /* multiple of stage+1 */
  if (Xmax== -1 && Ymax!= -1) Xmax = Ymax*ratio+1;
  else if (Ymax== -1 && Xmax!= -1) Ymax = Xmax/ratio+1;
}

int main(int argc, char* argv[]) {
  p4_initenv(&argc, argv);
  if (p4_get_my_id() == 0)
    p4_create_procgroup();
  if (p4_get_my_id() != 0) {
    slave();
  } else {
    master(argc, argv);
  }
  p4_wait_for_end();
}

int slave(void) {
  int i;
  int from, to = 0, type, len;
  int line, size;
  char* buf;
  int* setup = 0;
  double* range = 0;
  double ag,bg,ac,bc,az,bz,azl,s;
  int inner_loop_count;
  int limit,m,n,count,done = 0,t,inside;
  unsigned int bitmap;
  int iter;
  int* int_msg;

#ifdef VERBOSE
  if (verbose) {
    printf("hello from slave %d\n", p4_get_my_id());
    fflush(stdout);
  }
#endif

  from = 0;
  type = SETUP;

#ifdef VERBOSE
  if (verbose) {
    printf("slave %d waiting for stage 1 setup\n", p4_get_my_id());
    fflush(stdout);
  }
#endif

  setup = 0;
  p4_recv(&type, &from, (char**)&setup, &len);
#ifdef VERBOSE
  if (verbose) {
    printf("slave %d got stage 1 setup\n", p4_get_my_id());
    fflush(stdout);
  }
#endif

  Xmax = setup[0];  Ymax = setup[1];  iter = setup[2];
  p4_msg_free(setup);
  size = (Xmax + 7) / 8;
  buf = (char*) p4_malloc( size );

  from = 0;
  type = SETUP;

#ifdef VERBOSE
  if (verbose) {
    printf("slave %d waiting for stage 2 setup\n", p4_get_my_id());
    fflush(stdout);
  }
#endif

  range = 0;
  p4_recv(&type, &from, (char**)&range, &len);
#ifdef VERBOSE
  if (verbose) {
    printf("slave %d got stage 2 setup\n", p4_get_my_id());
    fflush(stdout);
  }
#endif

  a = range[0];  b = range[1];  as = range[2];  bs = range[3];
  p4_msg_free(range);

  ag = as/(Xmax+1.0);
  bg = bs/(Ymax+1.0);

  inner_loop_count=0;

  while ( !done ) {
    from = ANY;
    type = ANY;

#ifdef VERBOSE
    if (verbose) {
      printf("slave %d waiting for line\n", p4_get_my_id());
      fflush(stdout);
    }
#endif
    int_msg = 0;
    p4_recv(&type, &from, (char**)&int_msg, &len);
    line = *int_msg;
    p4_msg_free(int_msg);
#ifdef VERBOSE
    if (verbose) {
      printf("slave %d computing line %d\n", p4_get_my_id(), line);
      fflush(stdout);
    }
#endif
    if (line == -1) {
#ifdef REPORT_COUNT
      printf("%d: count = %d\n", p4_get_my_id(), inner_loop_count);
#endif
      done = 1;
      continue;
    }

    /* do work */

    bc = b + line * bg;
    ac = a;

    for (t=0 ; t<Xmax ; t+=16) {
      bitmap = 0;
      for (n=t; n<=t+15; n++) {
        az=0;
        bz=0;
        s=0;
        for (limit=iter ; limit; limit--) {
          azl=(az*az)-(bz*bz);
          bz=((double)2.0*az*bz)+bc;
          az=azl+ac;
          s=(az*az)+(bz*bz);
	  inside = (s<(double)4.0);
          if (!inside) break;
        }
#ifdef COUNT
	inner_loop_count += (iter - 1);
#endif
#ifndef TIMING
	bitmap = bitmap >> 1;
	if (inside) 
	  bitmap = bitmap | 32768;
#endif
        ac+=ag;
      }
#ifndef TIMING
      buf[t / 8] = bitmap & 255;
      buf[t / 8 + 1] = bitmap >> 8;
#endif
    }
    p4_send(RESULT, to, (char*)buf, size);
  }
}

int master(int argc, char* argv[]) {
#ifdef VERBOSE
  if (verbose) {
    printf("hello from master\n");
  }
#endif

  switches(argc,argv);
  argc -= nswitches;
  if (argc == 1) str[0] = '\0';
  else if (argv[1][0]=='0') str[0] = '\0';
  else strcpy(str,argv[1]);
  setupvars();
  if (!timing) {
    out = fopen(outfile,"w");
    if ( out == 0 ) {
      fprintf(stderr, "failed to open file %s\n", outfile);
      exit(1);
    }
    header();
  }
  area();
  if (user_iter != -1) {
    iter = user_iter;
  }
  printf("calculating %d x %d mandlebrot with %d iterations\n", 
	 Xmax, Ymax, iter);
  parcalc();
  if (!timing) {
    footer();
    fclose(out);
  }
}

void switches(argc,argv) 
int argc;
char *argv[];
{
int i;
char *rest;

  Xmax= -1;
  Ymax= -1;
  if (argc >1) {
    for (i=argc; i>=1 && argv[i-1][0]=='-'; i--) {
      rest = argv[i-1];
      rest++; rest++;
      nswitches++;
      switch (argv[i-1][1]) {
      case 'i':
	user_iter = atoi(rest);
	break;
      case 'x':
	Xmax = atoi(rest);
	break;
      case 'y':
	Ymax = atoi(rest);
	break;
      case 'v':
	verbose = 1;
	break;
      case 't':
	timing = 1;
	break;
      case 'o':
	outfile = rest;
	break;
      }
    }
  }
  if (Xmax== -1 && Ymax== -1) Xmax=64;
  if (Xmax != -1) Xmax = (Xmax + 7) & ~7; /* multiple of 8 */
  if (Xmax== -1 && Ymax!= -1) Xmax = Ymax*ratio+1;
  else if (Ymax== -1 && Xmax!= -1) Ymax = Xmax/ratio+1;
}

void header() {
  int HexWords=0;

  if (Xmax/8*8 == Xmax) HexWords=Xmax/8;
  else HexWords=Xmax/8+1;

  fprintf(out,"#define theicon_width %d\n",Xmax);
  fprintf(out,"#define theicon_height %d\n",Ymax);
  fprintf(out,"#define theicon_x_hot -1\n");
  fprintf(out,"#define theicon_y_hot -1\n");
  fprintf(out,"static char theicon_bits[] = {\n");
}

void footer() {
  fprintf(out,"}\n");
}

void area() {
int i=0;
/*  a= -2.568; 
  b= -1.25;
  as= 3.636; 
  bs= 2.5; */

  a= -2.357; 
  b= -1.25;
  as= 3.333; 
  bs= 2.5; 
  while (str[i]!=0) {
    switch (str[i]) {
      case '2':a=a+as/3; break;
      case '3':a=a+2*as/3; break;
      case '4':b=b+bs/3; break;
      case '5':a=a+as/3; b=b+bs/3; break;
      case '6':a=a+2*as/3; b=b+bs/3; break;
      case '7':b=b+2*bs/3; break;
      case '8':a=a+as/3; b=b+2*bs/3; break;
      case '9':a=a+2*as/3; b=b+2*bs/3; break;
       default:break;
    }
    as/=3;
    bs/=3;
    i++;
  }
  iter=autoiter[i];
}

struct told {
  int size;
  int queue[BUFFER];
};

void queue_init(struct told* q) {
  int i;
  q->size = 0;
  for (i = 0; i < BUFFER; i++) {
    q->queue[i] = -128;
  }
}

void queue_add(struct told* q, int v) {
  q->queue[q->size++] = v;
}

int queue_head(struct told* q) {
  int ret, i;
  ret = q->queue[0];
  for (i = 1; i < q->size; i++) {
    q->queue[i-1] = q->queue[i];
  }
  q->size--;
  return ret;
}

char* result;

void parcalc() {
  int i, j, linec;
  int last = p4_num_total_slaves();
  int was, from, to, tosend, todo, type, len, line, quit = -1;
  int start_time, end_time, elapsed_time;
  int setup[5];
  double range[4];
  int size = (Xmax + 7) / 8;
  char** lines;
  struct told* work;
  int* int_msg = 0;

  lines = (char**) p4_malloc(Ymax * sizeof(char*));

  work = (struct told*) p4_malloc((last + 1) * sizeof(struct told));

  for (i = 1; i <= last; i++) {
    queue_init(&work[i]);
  }

  setup[0] = Xmax;  setup[1] = Ymax;  setup[2] = iter;

  range[0] = a;  range[1] = b;  range[2] = as;  range[3] = bs;

  tosend = Ymax;

  start_time = p4_ustimer();
  line = 0;
  for (i = 1; i <= last; i++) {
#ifdef VERBOSE
    if (verbose) {
      printf("sending setup 1 to %d\n", i);
      fflush(stdout);
    }
#endif
    p4_send(SETUP, i, (char*)setup, 3 * sizeof(int));
#ifdef VERBOSE
    if (verbose) {
      printf("sending setup 2 to %d\n", i);
      fflush(stdout);
    }
#endif
    p4_send(SETUP, i, (char*)range, 4 * sizeof(double));
#ifdef VERBOSE
    if (verbose) {
      printf("telling slave %d to compute line %d\n", i, line);
      fflush(stdout);
    }
#endif
    queue_add(&work[i], line);
    p4_send(WORK, i, (char*)&line, sizeof(int));
    tosend--;
    line++;
  }
  
  for (j=1; j < BUFFER; j++) {
    for (i=1; i<=last && tosend; i++) {
#ifdef VERBOSE
      if (verbose) {
	printf("telling slave %d to compute line %d\n", i, line);
	fflush(stdout);
      }
#endif
      queue_add(&work[i], line);
      p4_send(WORK, i, (char*)&line, sizeof(int));
      line++;
      tosend--;
    }
  }
  todo = Ymax;

  while (tosend) {
    from = ANY;
    type = RESULT;
    
#ifdef VERBOSE
    if (verbose) {
      printf("main waiting for result from any slave\n");
      fflush(stdout);
    }
#endif
    result = 0;
    p4_recv(&type, &from, (char**)&result, &len);
    if (from < 1) {
      printf("cant tell where message 1 came from %d\n", from);
      fflush(stdout);
      from = 1;
    }
    was = queue_head(&work[from]);
#ifdef VERBOSE
    if (verbose) {
      printf("slave %d sent back result for line %d\n", from, was);
      fflush(stdout);
    }
#endif

    lines[was] = (char*)p4_malloc(size);
    memcpy(lines[was], result, size);
    p4_msg_free(result);
    for (;  work[from].size < BUFFER && tosend; ) {
#ifdef VERBOSE
      if (verbose) {
	printf("telling slave %d to compute line %d\n", from, line);
	fflush(stdout);
      }
#endif
      queue_add(&work[from], line);
      p4_send(WORK, from, (char*)&line, sizeof(int));
      line++; 
      todo--;
      tosend--;
    }
  }
  
  while (todo) {
    from = ANY;
    type = RESULT;

    result = 0;
    p4_recv(&type, &from, (char**)&result, &len);
    if (from < 1) {
      printf("cant tell where message 2 came from %d\n", from);
      fflush(stdout);
      from = 1;
    }
    was = queue_head(&work[from]);
#ifdef VERBOSE
    if (verbose) {
      printf("slave %d sent back result for line %d\n", from, was);
      fflush(stdout);
    }
#endif
    lines[was] = (char*)p4_malloc(size);
    memcpy(lines[was], result, size);
    p4_msg_free(result);
    p4_send(WORK, from, (char*)&quit, sizeof(int));
    todo--;
  }
  
  end_time = p4_ustimer();

  elapsed_time = end_time - start_time;

  printf("took %d ms\n", elapsed_time / 1000);

  if (!timing) {
    linec = 15;
    for (i = 0; i < Ymax; i++) {
      for (j = 0; j < size; j++) {
	fprintf(out,"0x%02x,",lines[i][j]);
	if (linec == 0) {
	  fprintf(out, "\n");
	  linec = 15;
	} else {
	  linec--;
	}
      }
    }
    printf("done writing file\n");
  }
}
