#include "p4.h"
#include "p4_sys.h"

/*
  Defining REDIRECT_OUTPUT allows you to redirect output from
  processes as follows.
  Defining OUT_TO_TERM tty allows you to redirect the
  output from various processes to separate windows.
  */

#if defined(REDIRECT_OUTPUT)
#if defined(OUT_TO_TERM)
#define P4_OUTFILE "/dev/ttyp4"
#else
#define P4_OUTFILE "/u/rbutler/p4/tests/out"
#endif
#endif

static int rm_num;

int rm_start(argc, argv)
     int *argc;
     char **argv;
{
  int bm_fd, bm_port;
  char *s;
#if defined(TRANSPUTER)
  int RootNodeBM;
  char bm_host[HOSTNAME_LEN];
#else
  char *bm_host;
#endif
  int externalBM = 1;
  extern char whoami[];
  struct bm_rm_msg msg;
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
  int rc;
  struct net_initial_handshake hs;
#endif
#if defined(IPSC860)  ||  defined(CM5)
  int i;
#endif
  int type, i, numslaves, numinpt;
  struct p4_global_data *g;
  char outfile[128];
  P4BOOL subnet_comms_only = FALSE;
#if defined(TRANSPUTER)
  Channel *fromRoot = ((Channel**) get_param( CONFIG_smallIn ))[ 0 ];
  Channel *toRoot = ((Channel**) get_param( CONFIG_smallOut ))[ 0 ];
#endif
  
#if !defined(TRANSPUTER)
  trap_sig_errs(); 		/* Errors can happen any time */
#endif
  
#if defined(TRANSPUTER)
  argc = argc; argv = argv;	/* keep compiler quiet */
#endif

  sprintf(whoami, "rm_%d", getpid());
  p4_dprintfl(20,"remote master starting\n");
  
#if defined(REDIRECT_OUTPUT)
  freopen(P4_OUTFILE, "w", stdout);
  freopen(P4_OUTFILE, "w", stderr);
#endif
  
#if defined(TRANSPUTER)
   /**********************************************************************\
   *                                                                      *
   * Can not use argv on remote processors on transputer, use             *
   * transputer comms with root transputer instead                        *
   *                                                                      *
   \**********************************************************************/
  RootNodeBM = ChanInInt( fromRoot );
  
  if ( !RootNodeBM ) {
    i = ChanInInt( fromRoot );
    ChanIn( fromRoot, bm_host, i );
    bm_port = ChanInInt( fromRoot );
  } else {
    bm_port = 0;
    externalBM = 0;
  }
#else
  if (*argc < 4)
    p4_error("Invalid arguments to remote_master", *argc);
  bm_host = argv[1];
  bm_port = atoi(argv[2]);
#endif
  
  if ( externalBM ) {
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
    bm_fd = net_conn_to_listener(bm_host, bm_port, 5);
    if (bm_fd < 0)
      p4_error("rm_start: net_conn_to_listener failed", bm_port);
    net_recv(bm_fd, &hs, sizeof(hs));
    hs.pid = (int) htonl(getpid());
    net_send(bm_fd, &hs, sizeof(hs), FALSE);
#endif
  } else {
    bm_fd = CHANNEL;  /* no need to handshake with channels */
  }
#   ifdef SYSV_IPC
  sysv_num_shmids = 0;
  sysv_shmid[0]  = -1;
  sysv_semid0 = init_sysv_semset(0);
#   endif
  
  /* Get the initialization information from the bm */
  
  if ( externalBM ){
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
    rc = net_recv(bm_fd, &msg, sizeof(msg));
    if (rc == PRECV_EOF)
      p4_error("rm_start: got EOF on net_recv", bm_fd);
    type = p4_n_to_i(msg.type);
    globmemsize = p4_n_to_i(msg.memsize);
    logging_flag = p4_n_to_i(msg.logging_flag);
    numslaves = p4_n_to_i(msg.numslaves);
    rm_num = p4_n_to_i(msg.rm_num);
    debug_level = p4_n_to_i(msg.debug_level);
    numinpt = p4_n_to_i(msg.numinproctab);
#endif
  } else {
#if defined(TRANSPUTER)
    ChanIn( fromRoot, &msg, sizeof( msg ) );
    type = msg.type;
    globmemsize = msg.memsize;
    logging_flag = msg.logging_flag;
    numslaves = msg.numslaves;
    rm_num = msg.rm_num;
    debug_level = msg.debug_level;
    numinpt = msg.numinproctab;
    subnet_comms_only = msg.port;
#endif
  }
  if (type != INITIAL_INFO)
    p4_error("rm_start: unknown type, expecting INITIAL_INFO, type=", type);
  if (strcmp(msg.version,p4_version()) != 0)
    {
      p4_dprintf("my version is %s\n",p4_version());
      p4_error("version does not match master \n",0);
    }
  
  if ((s = (char *) rindex(msg.pgm,'/'))  ==  NULL)
    {
      p4_dprintf("my fullpathname is invalid: %s\n",msg.pgm);
      p4_error("fullpathname is invalid \n",0);
    }
  else
    {
      *s = '\0';  /* chg to directory name only */
#ifndef TRANSPUTER
      chdir(msg.pgm);
#endif
    }
  
  if (logging_flag)
    ALOG_ENABLE;
  else
    ALOG_DISABLE;
  
  MD_initmem(globmemsize);
  alloc_global();  /* sets p4_global */
  g = p4_global;
#ifdef TRANSPUTER
  MainLocal = alloc_local_rm();
#else
  p4_local = alloc_local_rm();
#endif
  g->local_communication_only = FALSE;
  g->subnet_communication_only = subnet_comms_only;
  g->num_in_proctable = numinpt;
  strcpy(outfile, msg.outfile);
  strcpy(p4_global->application_id, msg.application_id);
  p4_dprintfl(90, "got numslaves=%d outfile=%s rm_num=%d dbglvl=%d appid=%s\n",
	      numslaves, outfile, rm_num, debug_level, msg.application_id);
  
  MD_initenv();
  usc_init();
  
  if (*outfile)
    {
      freopen(outfile, "w", stdout);
      freopen(outfile, "w", stderr);
    }
  
#if !defined(TRANSPUTER)
  SIGNAL_P4(LISTENER_ATTN_SIGNAL, handle_connection_interrupt);
#endif
  
  p4_lock(&g->slave_lock);
  create_rm_processes(numslaves, bm_fd);
  
  /* Grab the whole proc table from the bm */
  p4_dprintfl(90, "receiving proc table\n");
  receive_proc_table(bm_fd);
  
  /* let local slaves use proctable to identify themselves */
  p4_unlock(&g->slave_lock);
  
  sprintf(whoami, "rm_%d_%d", rm_num, getpid());
  p4_local->my_id = p4_get_my_id_from_proc();
  
  p4_global->low_cluster_id = 
    p4_local->my_id - p4_global->proctable[p4_local->my_id].slave_idx;
  p4_global->hi_cluster_id = 
    p4_global->low_cluster_id + p4_global->local_slave_count;
  
  setup_conntab();
  
  if (p4_local->conntab[0].type == CONN_REMOTE_SWITCH)
    {
      p4_local->conntab[0].switch_port = p4_global->proctable[0].switch_port;
      p4_local->conntab[0].port = bm_fd;
    }
  else if (p4_local->conntab[0].type == CONN_REMOTE_NON_EST)
    {
      p4_local->conntab[0].type = CONN_REMOTE_EST;
      p4_local->conntab[0].port = bm_fd;
      p4_local->conntab[0].same_data_rep =
	same_data_representation(p4_local->my_id,0);
    }
#if defined(TRANSPUTER)
  else if ( p4_local->conntab[0].type == CONN_CHANNEL ) 
    {
      /* ok */;
    }
#endif
  else
    {
      p4_error("rm_start: invalid conn type in conntab ",
	       p4_local->conntab[0].type);
    }
  sprintf(whoami, "p%d_%d", p4_get_my_id(), getpid());
  
  
#if defined(IPSC860)  ||  defined(CM5)
  for (i = 1; i < numslaves; i++)
    {
#       if defined(IPSC860)
      csend((long) INITIAL_INFO, &msg, (long) sizeof(struct bm_rm_msg), 
	    (long) i, (long) NODE_PID);
      csend((long) INITIAL_INFO, p4_global->proctable, 
	    (long) sizeof(p4_global->proctable), (long) i, (long) NODE_PID);
#       endif
#       if defined(CM5)
      CMMD_send_noblock(i, INITIAL_INFO,  &msg,sizeof(struct bm_rm_msg));
      CMMD_send_noblock(i, INITIAL_INFO,  p4_global->proctable, 
			sizeof(p4_global->proctable));
#       endif
    }
#endif
  
  /* 
    sync with local slaves thus insuring that they have the proctable before 
    syncing with bm (this keeps bm and its slaves  from interrupting the local 
    processes too early; then re-sync with local slaves (thus permitting them to 
    interrupt remotes) )
    */
  p4_barrier(&(p4_global->cluster_barrier),p4_num_cluster_ids());
  
  if ( externalBM ) {
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
    msg.type = p4_i_to_n(SYNC_MSG);
    net_send(bm_fd, &msg, sizeof(msg), FALSE);
#endif
  } else {
    msg.type = SYNC_MSG;
    ChanOut( toRoot, &msg, sizeof( msg ) );
  }
  msg.type = -1;  /* reset to verify type received next */
  if ( externalBM ) {
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
    rc = net_recv(bm_fd, &msg, sizeof(msg));
    type = p4_n_to_i(msg.type);
#endif
  } else {
    ChanIn( fromRoot, &msg, sizeof( msg ) );
    type = msg.type;
  }
  if (type != SYNC_MSG)
    p4_error("rm_start: unknown type, expecting SYNC_MSG, type=", type);
  
#if defined(TRANSPUTER)
  t_comms_init();
#endif
  p4_barrier(&(p4_global->cluster_barrier),p4_num_cluster_ids());
  
  return(0);
}

#if defined(TRANSPUTER)

static int slave_idx, listener_port;
#if defined(SOCKET_SERVER)
static int listener_fd;
#endif

void create_slave_for_rm( void ) {
  /* In the slave process */

  sprintf(whoami, "rm_s_%d_%d_%d", rm_num, slave_idx, getpid());

#ifdef TRANSPUTER
  alloc_local_slave();
#else
  p4_local = alloc_local_slave();
#endif

  /* hang for a valid proctable */
  p4_lock(&p4_global->slave_lock);
  p4_unlock(&p4_global->slave_lock);

  p4_local->my_id = p4_get_my_id_from_proc();
  sprintf(whoami, "p%d_%d", p4_get_my_id(), getpid());
  setup_conntab();
  usc_init();

  /* 
    sync with local master twice: once to make sure all slaves 
    have got proctable, and second after the master has synced bm
    */
  p4_barrier(&(p4_global->cluster_barrier),p4_num_cluster_ids());
  p4_barrier(&(p4_global->cluster_barrier),p4_num_cluster_ids());
  
  p4_dprintfl(20, "remote process starting\n");
  ALOG_SETUP(p4_local->my_id,ALOG_TRUNCATE);
  ALOG_LOG(p4_local->my_id,BEGIN_USER,0,"");
  slave();
  ALOG_LOG(p4_local->my_id,END_USER,0,"");
  ALOG_OUTPUT;
  p4_dprintfl(20, "remote process exiting\n");
}

#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
void create_listener_for_rm( void ) {
  /* Inside listener */
  sprintf(whoami, "rm_l_%d_%d", rm_num, getpid());
  p4_dprintfl(70, "inside listener pid %d\n", getpid());
#ifdef TRANSPUTER
  alloc_local_listener();
#else
  p4_local = alloc_local_listener();
#endif
  listener();
}
#endif
#endif

P4VOID create_rm_processes(nslaves, bm_fd)
int nslaves;
int bm_fd;
{
    struct p4_global_data *g = p4_global;
    int slave_pid, listener_pid;
#if !defined(TRANSPUTER)
    int end_1, end_2;
    int slave_idx, listener_port, listener_fd;
#endif
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
    struct listener_data *l;
#endif
    char rm_host[100];
    int rm_switch_port;
    struct bm_rm_msg bm_msg;
#if defined(TRANSPUTER)
    Channel *toRoot = ((Channel**) get_param( CONFIG_smallOut ))[ 0 ];
#endif
    P4BOOL use_channels;
#   if !defined(IPSC860)  &&  !defined(CM5)
    if (nslaves > P4_MAX_MSG_QUEUES)
	p4_error("create_rm_processes: more slaves than msg queues \n", nslaves);
#   endif

#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
    if ( !p4_global->subnet_communication_only ) {
      /*
       * Allocate the listener's local data area, since this process will
       * eventually become the listener.
       */

      l = listener_info = alloc_listener_info();
      
      net_setup_anon_listener(10, &listener_port, &listener_fd);

      p4_dprintfl(70, "created listener on port %d fd %d\n", listener_port,
		  listener_fd);

      /* Send off the listener info to the bm */
      bm_msg.type = p4_i_to_n(REMOTE_LISTENER_INFO);
      bm_msg.port = p4_i_to_n(listener_port);
      net_send(bm_fd, &bm_msg, sizeof(struct bm_rm_msg), FALSE);
    }
#endif

    rm_host[0] = '\0';
    get_qualified_hostname(rm_host);
    rm_switch_port = getswport(rm_host);

    /* Send my info to the bm */
    strcpy(bm_msg.host_name,rm_host);
    strcpy(bm_msg.machine_type,P4_MACHINE_TYPE);
    use_channels = FALSE;
#if defined(TRANSPUTER)
    if ( bm_fd == CHANNEL ) {
      use_channels = TRUE;
      bm_msg.type = REMOTE_MASTER_INFO;
      bm_msg.slave_idx = 0;
      bm_msg.slave_pid = getpid();
      bm_msg.switch_port = rm_switch_port;
      bm_msg.network = my_network_no;
      bm_msg.node = MYNODE();
      ChanOut( toRoot, &bm_msg, sizeof( struct bm_rm_msg ) );
    }
#endif
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
    if ( !use_channels ) {
      bm_msg.type = p4_i_to_n(REMOTE_MASTER_INFO);
      bm_msg.slave_idx = p4_i_to_n(0);
      bm_msg.slave_pid = p4_i_to_n(getpid());
      bm_msg.switch_port = p4_i_to_n(rm_switch_port);
      bm_msg.network = p4_i_to_n(my_network_no);
      bm_msg.node = p4_i_to_n( MYNODE() );
      net_send(bm_fd, &bm_msg, sizeof(struct bm_rm_msg), FALSE);
    }
#endif

    g->local_slave_count = 0;

#   ifdef TCMP
    tcmp_init(NULL,p4_get_my_cluster_id(),shmem_getclunid());
#   endif

#   if defined(IPSC860)  ||  defined(CM5)
    for (slave_idx = 1; slave_idx <= nslaves - 1; slave_idx++)
    {
#       if defined(IPSC860)
	crecv(INITIAL_INFO, &bm_msg, (long) sizeof(struct bm_rm_msg));
#       endif
#       if defined(CM5)
        CMMD_receive(CMMD_ANY_NODE, INITIAL_INFO, (void *) &bm_msg, 
                     sizeof(struct bm_rm_msg));
#       endif
	net_send(bm_fd, &bm_msg, sizeof(struct bm_rm_msg), FALSE);
	g->local_slave_count++;
    }
#   else

#if !defined(TRANSPUTER)
    get_pipe(&end_1, &end_2);
#endif
    for (slave_idx = 1; slave_idx <= nslaves - 1; slave_idx++)
    {
	p4_dprintfl(20,"remote master creating local slave %d\n",slave_idx);
		    
#if defined(TRANSPUTER)
	slave_pid = proc_fork_p4( create_slave_for_rm );
#else
	slave_pid = fork_p4();
	if (slave_pid)
	    p4_dprintfl(10,"remote master created local slave %d\n",slave_idx);
	if (slave_pid == 0)
	{
	    /* In the slave process */

	    sprintf(whoami, "rm_s_%d_%d_%d", rm_num, slave_idx, getpid());

	    close(listener_fd);
	    p4_local = alloc_local_slave();
	    p4_local->listener_fd = end_1;
	    close(end_2);

            SIGNAL_P4(LISTENER_ATTN_SIGNAL, handle_connection_interrupt);

	    /* hang for a valid proctable */
	    p4_lock(&g->slave_lock);
	    p4_unlock(&g->slave_lock);

	    p4_local->my_id = p4_get_my_id_from_proc();
	    sprintf(whoami, "p%d_%d", p4_get_my_id(), getpid());
	    setup_conntab();
	    usc_init();

#           ifdef TCMP
            tcmp_init(NULL,p4_get_my_cluster_id(),shmem_getclunid());
#           endif

	    /* 
	       sync with local master twice: once to make sure all slaves 
	       have got proctable, and second after the master has synced bm
	    */
	    p4_barrier(&(p4_global->cluster_barrier),p4_num_cluster_ids());
	    p4_barrier(&(p4_global->cluster_barrier),p4_num_cluster_ids());

	    p4_dprintfl(20, "remote process starting\n");
            ALOG_SETUP(p4_local->my_id,ALOG_TRUNCATE);
            ALOG_LOG(p4_local->my_id,BEGIN_USER,0,"");
	    slave();
            ALOG_LOG(p4_local->my_id,END_USER,0,"");
            ALOG_OUTPUT;
	    p4_dprintfl(20, "remote process exiting\n");

	    exit(0);
	}
#endif
	/* Send off the slave info to the bm */
	strcpy(bm_msg.machine_type,P4_MACHINE_TYPE);
	use_channels = FALSE;
#if defined(TRANSPUTER)
	if ( bm_fd == CHANNEL ) {
	  use_channels = TRUE;
	  bm_msg.type = REMOTE_SLAVE_INFO;
	  bm_msg.slave_idx = slave_idx;
	  bm_msg.slave_pid = slave_pid;
	  bm_msg.switch_port = rm_switch_port;
	  bm_msg.network = my_network_no;
	  bm_msg.node = MYNODE();
	  ChanOut( toRoot, &bm_msg, sizeof( struct bm_rm_msg ) );
	} 
#endif
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
	if ( !use_channels ) {
	  bm_msg.type = p4_i_to_n(REMOTE_SLAVE_INFO);
	  bm_msg.slave_idx = p4_i_to_n(slave_idx);
	  bm_msg.slave_pid = p4_i_to_n(slave_pid);
	  bm_msg.switch_port = p4_i_to_n(rm_switch_port);
	  bm_msg.network = p4_i_to_n(my_network_no);
	  bm_msg.node = p4_i_to_n( MYNODE() );
	  net_send(bm_fd, &bm_msg, sizeof(struct bm_rm_msg), FALSE);
	}
#endif
	g->local_slave_count++;
    }
#endif

    /* Send the end message to the bm */
    use_channels = FALSE;
#if defined(TRANSPUTER)
    if ( bm_fd == CHANNEL ) {
      use_channels = TRUE;
      bm_msg.type = REMOTE_SLAVE_INFO_END;
      ChanOut( toRoot, &bm_msg, sizeof( struct bm_rm_msg ) );
    }
#endif
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
    if ( !use_channels ) {
      bm_msg.type = p4_i_to_n(REMOTE_SLAVE_INFO_END);
      net_send(bm_fd, &bm_msg, sizeof(struct bm_rm_msg), FALSE);
    }
#endif
    /*
     * Done creating slaves. Now fork off the listener .. we've already
     * created the socket and bound a port to it
     */

#   if !defined(IPSC860)  &&  !defined(CM5)
    g->listener_port = listener_port;
#if defined(TRANSPUTER)
#if defined(SOCKET_SERVER)
    if ( !p4_global->subnet_communication_only ) {
      listener_pid = proc_fork_p4( create_listener_for_rm );
    }
#endif
#else
    p4_local->listener_fd = end_1;
    listener_pid = fork_p4();
    if (listener_pid == 0)
    {
	/* Inside listener */
	sprintf(whoami, "rm_l_%d_%d", rm_num, getpid());
	p4_dprintfl(70, "inside listener pid %d\n", getpid());
	p4_local = alloc_local_listener();
	l->listening_fd = listener_fd;
	l->slave_fd = end_2;
	close(end_1);
	listener();
	exit(0);
    }
#endif

    /* Else we're still in the remote master */

    p4_dprintfl(70, "created listener pid %d\n", listener_pid);
    /* We need to close the fds from the listener setup */
# if !defined(TRANSPUTER)
    close(listener_fd);
    close(end_2);
#endif
    g->listener_pid = listener_pid;
#   endif
}


P4VOID receive_proc_table(bm_fd)
int bm_fd;
{
    P4BOOL done, use_channels;
    struct bm_rm_msg msg;
    int type;
    int port, unix_id, slave_idx, group_id;
    int switch_port, network, node;
#if defined(TRANSPUTER)
    Channel *fromRoot = ((Channel**) get_param( CONFIG_smallIn ))[ 0 ];
#endif

    p4_dprintfl(90, "receive_proc_table\n");
    for (done = FALSE; !done;)
    {
      use_channels = FALSE;
#if defined(TRANSPUTER)
        if ( bm_fd == CHANNEL ) {
	  use_channels = TRUE;
	  ChanIn( fromRoot, &msg, sizeof( msg ) );
	  type = msg.type;
	}
#endif
        if ( !use_channels ) {
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
	  if (net_recv(bm_fd, &msg, sizeof(msg)) == PRECV_EOF)
	    p4_error("recv_proc_table: got net_send_eof", bm_fd);
	  type = p4_n_to_i(msg.type);
#endif
	}
	switch (type)
	{
	  case PROC_TABLE_ENTRY:
	    if ( use_channels ) {
	      group_id = msg.group_id;
	      port = msg.port;
	      unix_id = msg.unix_id;
	      slave_idx = msg.slave_idx;
	      switch_port = msg.switch_port;
	      network = msg.network;
	      node = msg.node;
	    } 
#if !defined(TRANSPUTER) || defined(SOCKET_SERVER)
	    else {
	      group_id = p4_n_to_i(msg.group_id);
	      port = p4_n_to_i(msg.port);
	      unix_id = p4_n_to_i(msg.unix_id);
	      slave_idx = p4_n_to_i(msg.slave_idx);
	      switch_port = p4_n_to_i(msg.switch_port);
	      network = p4_n_to_i(msg.network);
	      node = p4_n_to_i(msg.node);
	    }
#endif
	    p4_dprintfl(90, "got entry gid=%d host=%s port=%d unix_id=%d slave_idx=%d switch_port=%d\n",
			group_id,msg.host_name,port,unix_id,slave_idx,switch_port);
	    /* remote master loading proctable from big master */
	    install_in_proctable(group_id, port, unix_id, msg.host_name,
				 slave_idx, msg.machine_type, switch_port,
				 network, node);
	    break;

	  case PROC_TABLE_END:
	    done = TRUE;
	    break;

	  default:
	    p4_dprintf("receive_proc_table: got invalid message type %d\n", type);
	    break;
	}
    }
}
