/**********************************************************************\
*                                                                      *
*                  Copyright (c) Exeter University,		       *
*                  Department of Computer Science,                     *
*                 (aba@uk.ac.ex.dcs) Adam Back, 1994                   *
*                                                                      *
*  GNU GPL (General Public License) Applies.                           *
*                                                                      *
\**********************************************************************/
#include "p4.h"
#include "p4_sys.h"
#include "t_comms.h"

/* #define ROTATE_BUFFER_HACK */

#define SERVER_STACK 1024

#if defined(T9000)
/* #define USE_VAR_MSGS */
#endif
int RDEBUG = 0;

/* #define DEBUG */
/* #define PDEBUG */

#if defined(USE_VAR_MSGS)
#define MAX_VAR_MSG 128
#endif

static void buffer_manager( Process*, Channel**, struct p4_buf* );
static void multiplexor( Process*, struct p4_buf*, Channel* );
static void demultiplexor( Process*, Channel* );

void OutOfMem( void );

/* start up synchronization vars */

Semaphore ready;

void t_comms_init() {
  int no_processors = p4_global->noProcs;
  int no_processes = p4_num_cluster_ids();
  Channel** smallIn = p4_global->smallIn;
  Channel **smallOut = p4_global->smallOut;
  Channel* actual_manager_chans;
  
  int *wspace;

  Process *muxor, *demuxor, *mgr;
  int i, j;

/* create buffers */

/* per process */

/* O/S buffers - process receive queue */

  p4_global->os_bufs = (struct p4_buf*)
    t_malloc( no_processes * sizeof( struct p4_buf ) );
  if ( p4_global->os_bufs == 0 ) OutOfMem();

/* per processor */

/* send queues */

  p4_global->small_send_q = (struct p4_buf*)
    t_malloc( no_processors * sizeof( struct p4_buf ) );
  if ( p4_global->small_send_q == 0 ) OutOfMem();

/* initialize */

/* buffers */

  for ( i = 0; i < no_processors; i++ ) {

/* send queues */

    p4_global->small_send_q[ i ].head = 0;
    p4_global->small_send_q[ i ].tail = 0;
    p4_lock_init( &p4_global->small_send_q[ i ].lock );
/* start out in a locked state */
    p4_lock( &p4_global->small_send_q[ i ].lock );
    ChanInit( &p4_global->small_send_q[ i ].active );

  }

/* O/S buffers */

  for ( i = 0; i < no_processes; i++ ) {
    p4_global->os_bufs[ i ].head = 0;
    p4_global->os_bufs[ i ].tail = 0;
    p4_lock_init( &p4_global->os_bufs[ i ].lock );
    p4_global->os_bufs[ i ].sleeping = 0;
    ChanInit( &p4_global->os_bufs[ i ].ack );
  }

  p4_global->manager_chans =
    (Channel***) t_malloc( (no_processors) * sizeof( Channel**) );
  if ( p4_global->manager_chans == 0 ) OutOfMem();

  actual_manager_chans = 
    (Channel*) t_malloc( no_processors * no_processes * sizeof( Channel ) );
  if ( actual_manager_chans == 0 ) OutOfMem();

  for ( i = 0; i < no_processors; i++ ) {
    p4_global->manager_chans[ i ] = 
      (Channel**) t_malloc( (no_processes+1) * sizeof( Channel* ) );
    if ( p4_global->manager_chans[ i ] == 0 ) OutOfMem();
    for ( j = 0; j < no_processes; j++ ) {
      ChanInit( &actual_manager_chans[ i * no_processes + j ] );
      p4_global->manager_chans[ i ][ j ] = 
	&actual_manager_chans[ i * no_processes + j ];
    }
    p4_global->manager_chans[ i ][ no_processes ] = 0;
  }

/* start multiplexors and buffer managers */
  for ( i = 0; i < no_processors; i++ ) {
    if ( p4_global->processor_no != i ) {

      muxor = (Process*)t_malloc( sizeof(Process) );
      if ( muxor == 0 ) exit(1);
      wspace = (int*)t_malloc( SERVER_STACK );
      if ( wspace == 0 ) exit(1);

      ProcInit( muxor, multiplexor, wspace, SERVER_STACK, 2, 
	        &p4_global->small_send_q[ i ], smallOut[ i ] );

      ProcRunHigh( muxor );

/* buffer manager */

      mgr = (Process*)t_malloc( sizeof(Process) );
      if ( mgr == 0 ) exit(1);
      wspace = (int*)t_malloc( SERVER_STACK );
      if ( wspace == 0 ) exit(1);

      ProcInit( mgr, buffer_manager, wspace, SERVER_STACK, 2, 
	        p4_global->manager_chans[ i ], &p4_global->small_send_q[ i ] );

      ProcRunHigh( mgr );
    }
  }

/* start demultiplexors */
  for ( i = 0; i < no_processors; i++ ) {
    if ( p4_global->processor_no != i ) {
      demuxor = (Process*)t_malloc( sizeof(Process) );
      if ( demuxor == 0 ) exit(1);
      wspace = (int*)t_malloc( SERVER_STACK );
      if ( wspace == 0 ) exit(1);

      ProcInit( demuxor, demultiplexor, wspace, SERVER_STACK, 1, 
	        smallIn[ i ] );

      ProcRunHigh( demuxor );

    }
  }

  for ( i = 0; i < no_processors; i++ ) {
    p4_unlock( &p4_global->small_send_q[ i ].lock );
  }
}

static void buffer_manager( Process* p, Channel** in, struct p4_buf* buf) {
  int from;
  int muxer_asleep = 0;
  struct p4_msg* msg;
  while ( 1 ) {
    from = ProcAltList( in );
    msg = (struct p4_msg*) ChanInInt( in[ from ] );
    p4_lock( &buf->lock );
    if ( buf->head == 0 ) {
      muxer_asleep = 1;
      buf->head = msg;
      buf->head->link = 0;
    } else {
      buf->tail->link = msg;
    }
    buf->tail = msg;
    p4_unlock( &buf->lock );
    if ( muxer_asleep ) {
      ChanOutChar( &buf->active, 'w' );
      muxer_asleep = 0;
    }
  }
}

static void multiplexor( Process *p, struct p4_buf* me, Channel *out ){
  struct p4_msg *msg;
  int sender_index;
  int go_to_sleep = 0;

  ChanInChar( &me->active );
  while ( 1 ) {

    msg = me->head;
    p4_lock( &me->lock );
    if ( me->head == me->tail ) {
      me->head = 0;
      go_to_sleep = 1;
    } else {
      me->head = me->head->link;
    }				/* remove message from send q */
    p4_unlock( &me->lock );

    if (RDEBUG)
      printf("multiplexor: sending msg size %d\n", msg->len);

#if defined(USE_VAR_MSGS)
    if ( msg->len >= MAX_VAR_MSG ) {
      ChanVOut( out, msg, sizeof( struct p4_msg ) - sizeof( char* ) );
      ChanOut( out, (char*) (&msg->msg), msg->len );
    } else {
      ChanVOut( out, msg, 
	        msg->len + sizeof( struct p4_msg ) - sizeof( char* ) );
    }
#else
    ChanOut( out, msg, sizeof( struct p4_msg ) - sizeof( char* ) );
    if ( msg->len > 0 )
      ChanOut( out, (char*) (&msg->msg), msg->len );
#endif

#if defined(PDEBUG)
    p4_dprintfl(11,"send daemon forwarded message from %d to %d\n",msg->from,msg->to );
#endif

    if ( msg->ack_req & P4_ALLOCED_MSG ) {
      free_p4_msg( msg );
    }

    if ( go_to_sleep ) {
      ChanInChar( &me->active );
      go_to_sleep = 0;
    }

  }
  p = p;
}

#if defined(USE_VAR_MSGS)

static void demultiplexor( Process *p, Channel *in ) {
  struct p4_msg* msg;
  struct p4_msg* header = 0;
  struct p4_msg* old = 0;
  struct p4_buf *os_buf;
  int proc_index, receiver_asleep;
  int user_buffer, combined;

  p = p;			/* keep compiler quiet */

#if defined(PDEBUG)
  p4_dprintfl(11,"demultiplexor started\n");
#endif

  while ( 1 ) {
    if ( header == 0 ) {
      header = (struct p4_msg*) alloc_p4_msg( MAX_VAR_MSG );
    }
    ChanVIn( in, header, 
	     MAX_VAR_MSG + sizeof( struct p4_msg ) - sizeof( char* ) );

#if defined(PDEBUG)
    p4_dprintfl(11,"demultiplexor header of msg arrived\n");
    p4_dprintfl(11,"system received message from %d for %d\n", header->from, header->to);
#endif

    proc_index = header->to - p4_global->low_cluster_id;
    if ( header->ack_req & P4_ACK_REPLY_MASK ) {

/* acknowledgment message */

/* wake up process waiting for ack */
#if defined(PDEBUG)
      p4_dprintfl(11, "demultiplexor forwarding msg ack\n");
#endif

      ChanOutChar( &p4_global->os_bufs[ proc_index ].ack, 'w' );

    } else {

/* ordinary message */

      combined = header->len <= MAX_VAR_MSG;

      os_buf = &p4_global->os_bufs[ proc_index ];
      p4_lock( &os_buf->lock );

      receiver_asleep = os_buf->sleeping;

      if (RDEBUG)
	printf("demultiplexor: received msg size %d\n", header->len);

      user_buffer = receiver_asleep && os_buf->buf && 
	header->from == os_buf->from && header->type == os_buf->type;

      if ( user_buffer ) {
	msg = (struct p4_msg*)
	  ((char*)os_buf->buf - sizeof(struct p4_msg) + sizeof(char*));
	if ( combined ) {
	  bcopy( header, msg, 
		 header->len + sizeof( struct p4_msg ) - sizeof( char* ) );
	} else {
	  bcopy( header, msg, 
		 sizeof( struct p4_msg ) - sizeof( char* ) );
	  ChanIn( in, (char*) &(msg->msg), header->len );
	}
      } else {
	if ( combined ) {
	  msg = header;
	  header = 0;
	  if (RDEBUG)
	    printf("demultiplexor: stealing header %d\n", msg->len);
	} else {
	  msg = (struct p4_msg*) alloc_p4_msg( header->len );
	  if ( msg == 0 ) OutOfMem();
	  if (RDEBUG)
	    printf("demultiplexor: allocating msg %d\n", header->len);
	  bcopy( header, msg, 
		 sizeof( struct p4_msg ) - sizeof( char* ) );
	  ChanIn( in, (char*) &(msg->msg), header->len );
	}
      }
#if defined(PDEBUG)
      p4_dprintfl(11,"demultiplexor received msg 0x%08x for %d from %d\n",msg,header->to, header->from);
#endif
      if ( os_buf->head != 0 ) {
	os_buf->tail->link = msg;
      } else {
	os_buf->head = msg;
      }
      os_buf->tail = msg;
      if ( receiver_asleep ) {
#if defined(PDEBUG)	
	p4_dprintfl(11,"receiver %d was sleeping, so waking\n", proc_index);
#endif
	ChanOutChar(&os_buf->ack,'w');
      } else {
	p4_unlock( &os_buf->lock );
      }
    }
  }
}

#else

static void demultiplexor( Process *p, Channel *in ) {
  struct p4_msg *msg, header;
  struct p4_buf *os_buf;
  int proc_index, receiver_asleep;

  p = p;			/* keep compiler quiet */

#if defined(PDEBUG)
  p4_dprintfl(11,"demultiplexor started\n");
#endif

  while ( 1 ) {
    ChanIn( in, &header, sizeof( struct p4_msg ) - sizeof( char* ) );

#if defined(PDEBUG)
    p4_dprintfl(11,"demultiplexor header of msg arrived\n");
    p4_dprintfl(11,"system received message from %d for %d\n", header.from, header.to);
#endif

    proc_index = header.to - p4_global->low_cluster_id;
    if ( header.ack_req & P4_ACK_REPLY_MASK ) {

/* acknowledgment message */

/* wake up process waiting for ack */
#if defined(PDEBUG)
      p4_dprintfl(11, "demultiplexor forwarding msg ack\n");
#endif

      ChanOutChar( &p4_global->os_bufs[ proc_index ].ack, 'w' );

    } else {

/* ordinary message */

      os_buf = &p4_global->os_bufs[ proc_index ];

      p4_lock( &os_buf->lock );

      receiver_asleep = os_buf->sleeping;

      if (RDEBUG)
	printf("demultiplexor: received msg size %d\n", header.len);

      if ( receiver_asleep && os_buf->buf && 
	  header.from == os_buf->from && header.type == os_buf->type ) {
	msg = (struct p4_msg*)
	  ((char*)os_buf->buf - sizeof(struct p4_msg) + sizeof(char*));
      } else {
	msg = (struct p4_msg*) alloc_p4_msg( header.len );
	if ( msg == 0 ) OutOfMem();
	if (RDEBUG)
	  printf("demultiplexor: allocating msg %d\n", header.len);
      }

      bcopy( &header, msg, sizeof( struct p4_msg ) - sizeof( char* ) );

      if ( header.len != 0 ){
#if defined(PDEBUG)
	p4_dprintfl(11,"msg = 0x%08x, &(msg->msg) = 0x%08x\n",msg,&(msg->msg));
#endif
	ChanIn( in, (char*) &(msg->msg), header.len );
      }
#if defined(PDEBUG)
      p4_dprintfl(11,"demultiplexor received msg 0x%08x for %d from %d\n",msg,header.to, header.from);
#endif
      if ( os_buf->head != 0 ) {
	os_buf->tail->link = msg;
      } else {
	os_buf->head = msg;
      }
      os_buf->tail = msg;
      if ( receiver_asleep ) {
#if defined(PDEBUG)	
	p4_dprintfl(11,"receiver %d was sleeping, so waking\n", proc_index);
#endif
	ChanOutChar(&os_buf->ack,'w');
      } else {
	p4_unlock( &os_buf->lock );
      }
    }
  }
}

#endif

void chan_send( struct p4_msg *msg ) {
  int dest_processor, q_was_empty, sender_index, receiver_index;
  struct p4_buf *to;

  sender_index = msg->from - p4_global->low_cluster_id;
  dest_processor = p4_local->conntab[ msg->to ].port;

/* shared memory comms */

  if ( dest_processor == MYNODE() ) {
#if defined(PDEBUG)
    p4_dprintfl(11,"chan_send shared mem message sending\n");
#endif
    receiver_index = p4_global->proctable[ msg->to ].slave_idx;
    to = &p4_global->os_bufs[ receiver_index ];
    p4_lock( &to->lock );

    if ( msg->ack_req & P4_ACK_REPLY_MASK ) {
      ChanOutChar( &p4_global->os_bufs[ receiver_index ].ack, 'w' );
      p4_unlock( &to->lock );
    } else {
#if defined(PDEBUG)
      p4_dprintfl(11,"chan_send shared mem message sent\n");
#endif
      if (to->sleeping) {
#ifdef ROTATE_BUFFER_HACK
	if ( to->buf ) {
	  p4_msg_free( to->buf );  /* otherwise we have a memory leak */
	  to->buf = 0;		   /* note we are rotating buffers, but it */
	}			   /* saves a block copy */
#else
	if ( to->buf ) {
	  memcpy( to->buf - sizeof(struct p4_msg) + sizeof(char*),
		  msg,
		  msg->len + sizeof(struct p4_msg) - sizeof(char*) );
	  msg = (struct p4_msg*)to->buf;
	}
#endif
	if ( to->head == 0 ) {
	  to->head = msg;
	  to->head->link = 0;
	} else {
	  to->tail->link = msg;
	}
	to->tail = msg;
	ChanOutChar(&to->ack,'w');
      } else {
	if ( to->head == 0 ) {
	  to->head = msg;
	  to->head->link = 0;
	} else {
	  to->tail->link = msg;
	}
	to->tail = msg;
	p4_unlock( &to->lock );
      }
    }
  } else {

/* message passing */

#if defined(PDEBUG)
    p4_dprintfl(11,"chan_send remote message sending\n");
#endif

    ChanOutInt( 
      p4_global->manager_chans[ dest_processor ][ sender_index ],
      (int)msg );

#if defined(PDEBUG)
    p4_dprintfl(11,"chan_send remote message sent\n");
#endif
  }

/* acknowledge if reqd */

  if ( msg->ack_req & P4_ACK_REQ_MASK ) {
    ChanInChar( &p4_global->os_bufs[ sender_index ].ack );
  }
}

int chan_avail( void ) {
  int idx = p4_global->proctable[ p4_local->my_id ].slave_idx;
/*  printf( "my id is %d, and cluster id %d\n", p4_local->my_id, idx ); */
  return ( p4_global->os_bufs[ idx ].head != 0 );
}

struct p4_msg *chan_recv( int type, int from, char *usr_buf ){
  struct p4_msg *ack;
  struct p4_msg *msg;
  int idx = p4_global->proctable[ p4_local->my_id ].slave_idx;
  struct p4_buf *os_buf = &p4_global->os_bufs[ idx ];

#if defined(PDEBUG)
  p4_dprintfl(11, "channel_recv started\n" );
#endif

  p4_lock(&os_buf->lock);
  if (os_buf->head == 0) {
    os_buf->sleeping = 1;
    os_buf->buf = usr_buf;
    os_buf->type = type;
    os_buf->from = from;

#if defined(PDEBUG)
    p4_dprintfl(11, "channel_recv sleeping waiting for msg\n");
#endif
    p4_unlock(&os_buf->lock);
    ChanInChar(&os_buf->ack);

    os_buf->sleeping = 0;
#if defined(PDEBUG)
    p4_dprintfl(11, "channel_recv awoken on arrival of msg\n");
#endif
  }
  msg = os_buf->head;
#if defined(PDEBUG)
  p4_dprintfl(11, "channel_recv msg = 0x%08x\n",msg);  /*&&&&*/
#endif
#if defined(MDEBUG)
  if (msg == 0) {
    mflush( stdout );
  }
#endif
  if ( os_buf->head == os_buf->tail )
    os_buf->head = 0;
  else
    os_buf->head = os_buf->head->link;
  p4_unlock( &os_buf->lock );
  if ( msg->ack_req & P4_ACK_REQ_MASK ) {  /* send ack if required */
#if defined(PDEBUG)	
    p4_dprintfl(11, "channel_recv sending ack\n");  /*&&&&*/
#endif
    ack = (struct p4_msg*) alloc_p4_msg( 0 );
    ack->from = p4_get_my_id();
    ack->data_type = P4NOX;
    ack->len = 0;
    ack->to = msg->from;
    ack->ack_req = P4_ACK_REPLY_MASK | P4_ALLOCED_MSG;
    chan_send( ack );
  }
  return msg;
}

void OutOfMem( void ) {
  p4_error("out of memory\n",0);
  exit(1);
}
