/**********************************************************************\
*                                                                      *
*                  Copyright (c) Exeter University,		       *
*                  Department of Computer Science,                     *
*                 (aba@uk.ac.ex.dcs) Adam Back, 1994                   *
*                                                                      *
*  GNU GPL (General Public License) Applies.                           *
*                                                                      *
\**********************************************************************/
#include <stdiored.h>
#include <stdlib.h>
#include <channel.h>
#include <process.h>
#include <misc.h>
#include "cfs.h"

#define BUFFER_SIZE 4096

char *MBuffer, *DBuffer;

int noProcs;
Channel **serverIn, **serverOut;
Channel *hostServerIn, *hostServerOut;

void Muxer( void );
void Demuxer( void );

void Synch( Channel *c ){
  int i;
  for ( i = 0; i < 160; i++ ) {
    ChanOutChar( c, 0 );
  }
  ChanOutChar( c, 1 );
}

int socket_server_init( void ){
  Process *m, *d;
  noProcs =   *(int*)get_param( CONFIG_noProcs );
  serverIn =  (Channel**)get_param( CONFIG_serverIn );
  serverOut = (Channel**)get_param( CONFIG_serverOut );
  socketIn =  (Channel*)get_param( SERVER_socketIn );
  socketOut = (Channel*)get_param( SERVER_socketOut );
  Synch( serverOut );
  ChanOutInt( serverOut, noProcs );
  m = ProcAlloc( Muxer, 0, 0 );
  d = ProcAlloc( Demuxer, 0, 0 );
  ProcPar( m, d );
}

void Muxer( void ) {
  int from, len, to, who;
  MBuffer = (char*) malloc( BUFFER_SIZE );
  if ( MBuffer == 0 ) exit( 1 );
  socketIn[ noProcs ] = 0;
  while ( 1 ) {
    from = ProcAltList( socketIn );
    if ( from == -1 || from >= noProcs ) {
      exit( 1 );
    } else {
      who = ChanInInt( socketIn[ from ] );
      len = ChanInInt( socketIn[ from ] );
      ChanIn( socketIn[ from ], MBuffer, len );
      ChanOutInt( serverOut, from );
      ChanOutInt( serverOut, who );
      ChanOutInt( serverOut, len );
      ChanOut( serverOut, MBuffer, len );
    }
  }
}

void Demuxer( void ) {
  int len, to, who;
  DBuffer = (char*) malloc( BUFFER_SIZE );
  if ( DBuffer == 0 ) exit( 1 );
  while ( 1 ) {
    to = ChanInInt( serverIn );
    who = ChanInInt( serverIn );
    len = ChanInInt( serverIn );
    ChanIn( serverIn, DBuffer, len );
    ChanOutInt( socketOut[ to ], who );
    ChanOutInt( socketOut[ to ], len );
    ChanOut( socketOut[ to ], DBuffer, len );
  }
}
