#include <stdio.h>
#include <stdlib.h>

#include "t_amalloc.h"

const char* binr(unsigned int x) {
  static char str[20];
  int i;
  strcpy(str,"00000000000000000000000000000000");
  for (i=0; i<32 && x != 0; i++) {
    str[i] = (x & (1u << 31)) > 0 ? '1' : '0';
    x <<= 1;
  }
  return str;
}

#define bin(x) binr((unsigned int)(x))

void init(int size) {
  void* heap;
  heap = malloc(size);
  if (heap == 0) {
    fprintf(stderr, "malloc(%d) failed\n", size);
    exit(1);
  }
  t_malloc_init(heap, size);
}

#define malloc(x) t_malloc(x)
#define amalloc(x,y) t_amalloc(x,y)
#define free(x) t_free(x)

main() {
  void *p1, *p2, *p3, *p4;

  init(1024*1024);

  printf("testing malloc\n");

  report();

  p1 = malloc(23);
  printf("malloc(23) = 0x%08x %s\n",p1, bin(p1));
  p2 = malloc(25);
  printf("malloc(25) = 0x%08x %s\n",p2, bin(p2));
  p3 = malloc(27);
  printf("malloc(27) = 0x%08x %s\n",p3, bin(p3));
  p4 = malloc(29);
  printf("malloc(29) = 0x%08x %s\n",p4, bin(p4));

  report();

  free(p1);
  free(p4);
  free(p2);
  free(p3);

  printf("after freeing\n");
  report();

  printf("testing amalloc\n");

  p1 = amalloc(23, 4);
  printf("amalloc(23, 4) = 0x%08x %s\n", p1, bin(p1));
  report();
  p2 = amalloc(25, 8);
  printf("amalloc(25, 8) = 0x%08x %s\n", p2, bin(p2));
  report();
  p3 = amalloc(27, 16);
  printf("amalloc(27, 16) = 0x%08x %s\n", p3, bin(p3));
  report();
  p4 = amalloc(29, 2 << 15);
  printf("amalloc(29, 2 << 15) = 0x%08x %s\n", p4, bin(p4));

  report();

  free(p1);
  printf("free(0x%08x)\n", p1);
  report();
  free(p4);
  printf("free(0x%08x)\n", p4);
  report();
  free(p2);
  printf("free(0x%08x)\n", p2);
  report();
  free(p3);
  printf("free(0x%08x)\n", p3);
  report();
}
